import dash_ag_grid as dag
from dash import Dash, html
from . import utils


def test_cr001_custom_row_selectable(dash_duo):
    app = Dash(__name__)

    columnDefs = [
        {
            "field": "ticker",
            "headerCheckboxSelection": True,
            "checkboxSelection": True,
            "showDisabledCheckboxes": True,
        }
    ]
    grid = dag.AgGrid(
        id="grid",
        columnDefs=columnDefs,
        rowData=[{"ticker": "AAPL"}, {"ticker": "MSFT"}],
        dashGridOptions={
            "isRowSelectable": {
                "function": 'params.data.ticker == "AAPL" ? true: false'
            },
            "rowSelection": "multiple",
        },
    )

    app = Dash(__name__)

    app.layout = html.Div(grid)

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "AAPL")

    assert (
        dash_duo.find_element(
            '#grid [row-index="0"] [aria-colindex="1"] ' ".ag-selection-checkbox input"
        ).get_attribute("disabled")
        == None
    )
    assert dash_duo.find_element(
        '#grid [row-index="1"] [aria-colindex="1"] ' ".ag-selection-checkbox input"
    ).get_attribute("disabled")



from dash import Dash, html, Output, Input, no_update, State, ctx
import dash_ag_grid as dag
import plotly.express as px
import json
import pytest

from . import utils
from dash.testing.wait import until


df = px.data.election()
default_display_cols = ["district_id", "district", "winner"]


def test_us001_user_style(dash_duo):
    app = Dash(__name__)
    columnDefs = [
        {"headerName": "Make", "field": "make"},
        {"headerName": "Model", "field": "model"},
        {"headerName": "Price", "field": "price"},
    ]

    defaultColDef = {
        "initialWidth": 150,
        "sortable": True,
        "resizable": True,
        "filter": True,
    }

    rowData = [
        {"make": "Toyota", "model": "Celica", "price": 35000},
        {"make": "Ford", "model": "Mondeo", "price": 32000},
        {"make": "Porsche", "model": "Boxster", "price": 72000},
    ]

    app.layout = html.Div(
        [
            dag.AgGrid(
                id="grid",
                columnSize="autoSize",
                columnDefs=columnDefs,
                defaultColDef=defaultColDef,
                rowData=rowData,
                style={"height": "500px", "width": "500px"},
                className="ag-theme-alpine",
            ),
        ]
    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "Toyota")
    until(
        lambda: "height: 500px"
        in dash_duo.find_element("div.ag-theme-alpine").get_attribute("style"),
        timeout=3,
    )
    until(
        lambda: "width: 500px"
        in dash_duo.find_element("div.ag-theme-alpine").get_attribute("style"),
        timeout=3,
    )


@pytest.mark.parametrize("theme", ["alpine", "balham", "material", "quartz"])
def test_us002_legacy_themes(dash_duo, theme):
    app = Dash(
        __name__,
        external_stylesheets=[
            dag.themes.BASE,
            dag.themes.ALPINE,
            dag.themes.BALHAM,
            dag.themes.MATERIAL,
            dag.themes.QUARTZ,
        ],
    )

    columnDefs = [
        {"field": "name", "width": "500"},
    ]

    rowData = [
        {"name": "a"},
        {"name": "b"},
        {"name": "c"},
    ]

    app.layout = html.Div(
        [
            dag.AgGrid(
                id="grid",
                columnDefs=columnDefs,
                rowData=rowData,
                dashGridOptions={"theme": "legacy"},
                className=f"ag-theme-{theme}",
            ),
        ]
    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "a")

    # Test that the CSS files are actually loaded and applied

    # Base styles: assert that the grid height is <= 400px because an unstyled
    # grid is very "tall"
    until(
        lambda: dash_duo.find_element(".ag-root-wrapper").size["height"] <= 400,
        timeout=3,
        msg=f"Grid appears to be unstyled: height is too tall ({dash_duo.find_element('.ag-root-wrapper').size['height']}px)"
    )
    # Specific themes: Assert that cell headers are bold
    header_cell_text = dash_duo.find_element(".ag-header-cell-text")
    font_weight = header_cell_text.value_of_css_property("font-weight")
    assert font_weight in ["bold", "700", "600", "500",], "Grid appears to be unstyled: cell headers are not bold"

@pytest.mark.parametrize("theme", ["themeAlpine", "themeBalham", "themeMaterial", "themeQuartz"])
def test_us003_part_themes(dash_duo, theme):
    app = Dash(
        __name__
    )

    columnDefs = [
        {"field": "name", "width": "500"},
    ]

    rowData = [
        {"name": "a"},
        {"name": "b"},
        {"name": "c"},
    ]

    app.layout = html.Div(
        [
            dag.AgGrid(
                id="grid",
                columnDefs=columnDefs,
                rowData=rowData,
                dashGridOptions={'theme': {'function': f'customTheme({theme}, agGrid)'}},
            ),
        ]
    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "a")

    # Test that the CSS files are actually loaded and applied

    # Base styles: assert that the grid height is <= 400px because an unstyled
    # grid is very "tall"
    until(
        lambda: dash_duo.find_element(".ag-root-wrapper").size["height"] <= 400,
        timeout=3,
        msg=f"Grid appears to be unstyled: height is too tall ({dash_duo.find_element('.ag-root-wrapper').size['height']}px)"
    )

    # Specific themes: Assert that cell headers are bold
    header_cell_text = dash_duo.find_element(".ag-header-cell-text")
    font_weight = header_cell_text.value_of_css_property("font-weight")
    assert font_weight in ["bold", "700", "600", "500",], "Grid appears to be unstyled: cell headers are not bold"



from ._version import grid_version

_base_url = f"https://cdn.jsdelivr.net/npm/ag-grid-community@{grid_version}/styles"

BASE = f"{_base_url}/ag-grid.min.css"
ALPINE = f"{_base_url}/ag-theme-alpine.min.css"
BALHAM = f"{_base_url}/ag-theme-balham.min.css"
MATERIAL = f"{_base_url}/ag-theme-material.min.css"
QUARTZ = f"{_base_url}/ag-theme-quartz.min.css"



from dash import Dash, html, dcc, callback, Input, Output, Patch
import dash_ag_grid as dag
import pandas as pd

from . import utils


def test_fi006_quick_filter(dash_duo):
    app = Dash(__name__)
    df = pd.read_csv(
        "https://raw.githubusercontent.com/plotly/datasets/master/ag-grid/olympic-winners.csv"
    )

    columnDefs = [
        {"field": "athlete"},
        {"field": "age"},
    ]

    app.layout = html.Div(
        [
            dcc.Input(id="quick-filter-input", placeholder="filter..."),
            dag.AgGrid(
                id="quick-filter",
                rowData=df.to_dict("records"),
                columnDefs=columnDefs,
                dashGridOptions={
                    # allows to split by space (default) and ',' for the words to search
                    # 'michael phelps' works by default, now 'michael, phelps' or 'michael,phelps' also
                    "quickFilterParser": {"function": r"params.split(/[\s,]+/)"},
                    # allows to use regex in quick search, like 2[012] matching Age 20, 21 and 22
                    # defining in dashAgGridFunctions.js:
                    # dagfuncs.quickFilterMatcher = (quickFilterParts, rowQuickFilterAggregateText) = > {
                    #   return quickFilterParts.every(part= > rowQuickFilterAggregateText.match(part));
                    # }
                    "quickFilterMatcher": {"function": "quickFilterMatcher"}
                }
            ),
        ]
    )

    @callback(
        Output("quick-filter", "dashGridOptions"),
        Input("quick-filter-input", "value"),
        prevent_initial_call=True,
    )
    def update_filter(filter_value):
        new_grid_option = Patch()
        new_grid_option['quickFilterText'] = filter_value
        return new_grid_option

    # Tests ##################################################################

    dash_duo.start_server(app)
    grid = utils.Grid(dash_duo, "quick-filter")
    grid.wait_for_cell_text(0, 0, "Michael Phelps")

    quick_filter_input = dash_duo.find_element('#quick-filter-input')

    # test quickFilterParser
    quick_filter_input.send_keys("michael phelps")
    grid.wait_for_cell_text(0, 0, "Michael Phelps")
    dash_duo.clear_input(quick_filter_input)

    quick_filter_input.send_keys("michael, phelps")
    grid.wait_for_cell_text(0, 0, "Michael Phelps")
    dash_duo.clear_input(quick_filter_input)

    quick_filter_input.send_keys("michael,phelps")
    grid.wait_for_cell_text(0, 0, "Michael Phelps")
    dash_duo.clear_input(quick_filter_input)

    # test quickFilterMatcher
    quick_filter_input.send_keys("2[012]")
    grid.wait_for_cell_text(0, 1, "22")
    grid.wait_for_cell_text(1, 1, "21")
    grid.wait_for_cell_text(2, 1, "20")



import dash_ag_grid as dag
from dash import Dash, html, dcc, Output, Input, no_update, ctx
import requests
from . import utils


def test_sb001_selection_buttons(dash_duo):
    app = Dash(__name__)

    data = requests.get(
        r"https://www.ag-grid.com/example-assets/olympic-winners.json"
    ).json()

    columnDefs = [
        {"field": "athlete"},
        {"field": "age"},
        {"field": "country"},
        {"field": "year"},
        {"field": "date"},
        {"field": "sport"},
        {"field": "gold"},
        {"field": "silver"},
        {"field": "bronze"},
        {"field": "total"},
    ]

    app.layout = html.Div(
        [
            dcc.Markdown("This grid has single-select rows."),
            dag.AgGrid(
                id="grid",
                columnDefs=columnDefs,
                rowData=data,
                columnSize="sizeToFit",
                defaultColDef={
                    "resizable": True,
                    "sortable": True,
                    "filter": True,
                    "floatingFilter": True,
                },
                dashGridOptions={"rowSelection": "multiple"},
                persistence=True,
                persistence_type="session",
            ),
            html.Button(id="selectAll", children="Select All"),
            html.Button(id="selectAllFiltered", children="Select All Filtered"),
            html.Button(id="deselectAll", children="Deselect All"),
            html.Div(id="selections"),
        ],
        style={"margin": 20},
    )

    @app.callback(
        Output("grid", "selectAll"),
        Input("selectAll", "n_clicks"),
        Input("selectAllFiltered", "n_clicks"),
    )
    def setSelection(n, n2):
        if n:
            if ctx.triggered_id == "selectAll":
                return True
            else:
                return {"filtered": True}
        return no_update

    @app.callback(Output("grid", "deselectAll"), Input("deselectAll", "n_clicks"))
    def setSelection(n):
        if n:
            return True
        return no_update

    @app.callback(
        Output("selections", "children"),
        Input("grid", "selectedRows"),
    )
    def selected(selected):
        if selected:
            return len(selected)
        return "No selections"

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "Michael Phelps")
    grid.set_filter(0, "f")

    grid.wait_for_cell_text(0, 0, "Missy Franklin")

    dash_duo.find_element("#selectAll").click()
    dash_duo.wait_for_text_to_equal("#selections", "8618")
    dash_duo.find_element("#deselectAll").click()
    dash_duo.wait_for_text_to_equal("#selections", "No selections")
    dash_duo.find_element("#selectAllFiltered").click()
    dash_duo.wait_for_text_to_equal("#selections", "816")
    dash_duo.find_element("#deselectAll").click()
    dash_duo.wait_for_text_to_equal("#selections", "No selections")



import dash_ag_grid as dag
from dash import Dash, html, dcc, Output, Input, no_update
import requests
from . import utils
import pandas as pd
import time


def test_sr001_selected_rows(dash_duo):
    app = Dash(__name__)

    data = requests.get(
        r"https://www.ag-grid.com/example-assets/olympic-winners.json"
    ).json()

    columnDefs = [
        {"field": "athlete"},
        {"field": "age"},
        {"field": "country"},
        {"field": "year"},
        {"field": "date"},
        {"field": "sport"},
        {"field": "gold"},
        {"field": "silver"},
        {"field": "bronze"},
        {"field": "total"},
    ]

    app.layout = html.Div(
        [
            dcc.Markdown("This grid has single-select rows."),
            html.Div(id="selections-single-output"),
            dag.AgGrid(
                id="grid",
                columnDefs=columnDefs,
                rowData=data,
                columnSize="sizeToFit",
                defaultColDef={"resizable": True, "sortable": True, "filter": True},
                dashGridOptions={"rowSelection": "single"},
                persistence=True,
                persistence_type="session",
            ),
            html.Button(id="setSelection"),
        ],
        style={"margin": 20},
    )

    @app.callback(
        Output("selections-single-output", "children"),
        Input("grid", "selectedRows"),
    )
    def selected(selected):
        if selected:
            return f"You selected athlete: {selected[0]['athlete']}"
        return "No selections"

    @app.callback(
        Output("grid", "selectedRows"),
        Input("setSelection", "n_clicks"),
        prevent_initial_call=True,
    )
    def selected(n):
        if n > 0:
            return [
                {
                    "athlete": "Natalie Coughlin",
                    "age": 25,
                    "country": "United States",
                    "year": 2008,
                    "date": "24/08/2008",
                    "sport": "Swimming",
                    "gold": 1,
                    "silver": 2,
                    "bronze": 3,
                    "total": 6,
                }
            ]
        return no_update

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "Michael Phelps")

    grid.get_cell(0, 0).click()
    dash_duo.wait_for_text_to_equal(
        "#selections-single-output", "You selected athlete: Michael Phelps"
    )
    dash_duo.driver.refresh()

    grid.wait_for_cell_text(0, 0, "Michael Phelps")
    dash_duo.wait_for_text_to_equal(
        "#selections-single-output", "You selected athlete: Michael Phelps"
    )
    dash_duo.find_element("#setSelection").click()
    dash_duo.wait_for_text_to_equal(
        "#selections-single-output", "You selected athlete: Natalie Coughlin"
    )

def test_sr2_selected_rows_rowdata(dash_duo):
    app = Dash(__name__)

    # Placeholder DataFrame
    df = pd.DataFrame()

    app.layout = html.Div([
        html.Button('Update Table', id='update-button', n_clicks=0),
        html.Button('Select Two', id='select-two', n_clicks=0),
        dag.AgGrid(
            id='grid',
            columnSize="sizeToFit",
            columnDefs=[{'field': 'A'}],
            dashGridOptions={
                "rowHeight": None,
                "domLayout": "normal",
                "rowSelection": "multiple",
                "filter": True,
            },
            style={"maxHeight": "200px", "overflow": "auto"},
        ),
        html.Div(id='selected-row-info')  # Div to display selected row information
    ])

    @app.callback(
        Output('grid', 'columnDefs'),
        Output('grid', 'rowData'),
        Output('grid', 'selectedRows'),
        Input('update-button', 'n_clicks'),
        prevent_initial_call=True
    )
    def update_data(n_clicks):
        df = pd.DataFrame({
            "A": [1, 2, 3],
            "B": [4, 5, 6]
        })
        column_defs = [{"headerName": col, "field": col, "filter": "agTextColumnFilter"} for col in df.columns]
        row_data = df.to_dict("records")
        selected_rows = df.head(1).to_dict("records")
        return column_defs, row_data, selected_rows

    @app.callback(
        Output('selected-row-info', 'children'),
        Input('grid', 'selectedRows')
    )
    def update_selected_row_info(selected_rows):
        if isinstance(selected_rows, list):
            selected_row_info = f"Selected Row: {selected_rows[-1]}" if selected_rows else "No row selected"
            return selected_row_info
        return no_update

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_header_text(0, 'A')

    dash_duo.find_element('#update-button').click()

    grid.wait_for_cell_text(0, 0, "1")

    grid.get_cell(1, 0).click()
    dash_duo.wait_for_text_to_equal(
        "#selected-row-info", "Selected Row: {'A': 2, 'B': 5}"
    )

    dash_duo.find_element('#update-button').click()

    time.sleep(1) ## delay to make sure the selection sticks

    dash_duo.wait_for_text_to_equal(
        "#selected-row-info", "Selected Row: {'A': 1, 'B': 4}"
    )

    time.sleep(1)  ## delay to make sure the selection sticks

    assert 'ag-row-selected' in grid.get_row(0).get_attribute('class')

def test_sr3_selected_rows_modes(dash_duo):
    app = Dash(__name__)

    # Placeholder DataFrame
    df = pd.DataFrame()

    app.layout = html.Div([
        html.Button('Update Table', id='update-button', n_clicks=0),
        html.Button('Select Ids', id='select-ids', n_clicks=0),
        html.Button('Select Function', id='select-function', n_clicks=0),
        dag.AgGrid(
            id='grid',
            columnSize="sizeToFit",
            columnDefs=[{'field': 'A'}],
            dashGridOptions={
                "rowHeight": None,
                "domLayout": "normal",
                "rowSelection": "multiple",
                "filter": True,
            },
            style={"maxHeight": "200px", "overflow": "auto"},
            getRowId='params.data.A'
        ),
        html.Div(id='selected-row-info')  # Div to display selected row information
    ])

    @app.callback(
        Output('grid', 'columnDefs'),
        Output('grid', 'rowData'),
        Output('grid', 'selectedRows'),
        Input('update-button', 'n_clicks'),
        prevent_initial_call=True
    )
    def update_data(n_clicks):
        df = pd.DataFrame({
            "A": [1, 2, 3],
            "B": [4, 5, 6]
        })
        column_defs = [{"headerName": col, "field": col, "filter": "agTextColumnFilter"} for col in df.columns]
        row_data = df.to_dict("records")
        selected_rows = df.head(1).to_dict("records")
        return column_defs, row_data, selected_rows

    @app.callback(
        Output('grid', 'selectedRows', allow_duplicate=True),
        Input('select-ids', 'n_clicks'),
        prevent_initial_call=True
    )
    def select_ids(n_clicks):
        return {'ids': ['1', '2']}

    @app.callback(
        Output('grid', 'selectedRows', allow_duplicate=True),
        Input('select-function', 'n_clicks'),
        prevent_initial_call=True
    )
    def select_ids(n_clicks):
        return {'function': 'params.data.A > 2'}

    @app.callback(
        Output('selected-row-info', 'children'),
        Input('grid', 'selectedRows')
    )
    def update_selected_row_info(selected_rows):
        if isinstance(selected_rows, list):
            selected_row_info = f"Selected Row: {selected_rows[-1]}" if selected_rows else "No row selected"
            return selected_row_info
        return no_update

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_header_text(0, 'A')

    dash_duo.find_element('#update-button').click()

    grid.wait_for_cell_text(0, 0, "1")

    dash_duo.wait_for_text_to_equal(
        "#selected-row-info", "Selected Row: {'A': 1, 'B': 4}"
    )

    assert 'ag-row-selected' in grid.get_row(0).get_attribute('class')
    assert 'ag-row-selected' not in grid.get_row(1).get_attribute('class')
    assert 'ag-row-selected' not in grid.get_row(2).get_attribute('class')

    dash_duo.find_element('#select-ids').click()

    time.sleep(1)  ## delay to make sure the selection sticks

    dash_duo.wait_for_text_to_equal(
        "#selected-row-info", "Selected Row: {'A': 2, 'B': 5}"
    )

    time.sleep(1)  ## delay to make sure the selection sticks

    assert 'ag-row-selected' in grid.get_row(0).get_attribute('class')
    assert 'ag-row-selected' in grid.get_row(1).get_attribute('class')
    assert 'ag-row-selected' not in grid.get_row(2).get_attribute('class')

    dash_duo.find_element('#select-function').click()

    time.sleep(1)  ## delay to make sure the selection sticks

    dash_duo.wait_for_text_to_equal(
        "#selected-row-info", "Selected Row: {'A': 3, 'B': 6}"
    )

    time.sleep(1)  ## delay to make sure the selection sticks

    assert 'ag-row-selected' not in grid.get_row(0).get_attribute('class')
    assert 'ag-row-selected' not in grid.get_row(1).get_attribute('class')
    assert 'ag-row-selected' in grid.get_row(2).get_attribute('class')


"""
Nested tables.
"""

import dash_ag_grid as dag
import dash
from dash import Input, Output, html, dcc, Dash, no_update
from . import utils
import pandas as pd
from dash.testing.wait import until
import time


def test_is001_infinite_scroll(dash_duo):
    app = Dash(__name__)

    raw_data = {"id": [], "name": []}
    for i in range(0, 10000):
        raw_data["id"].append(i)
        raw_data["name"].append(f"{i * 3 % 5}-{i * 7 % 15}-{i % 8}")

    df = pd.DataFrame(data=raw_data)

    app.layout = html.Div(
        [
            dag.AgGrid(
                id="grid",
                columnSize="sizeToFit",
                columnDefs=[
                    {"field": "id", "filter": "agNumberColumnFilter"},
                    {"field": "name"},
                ],
                defaultColDef={
                    "sortable": True,
                    "filter": True,
                    "floatingFilter": True,
                },
                rowModelType="infinite",
                dashGridOptions={
                    # The number of rows rendered outside the viewable area the grid renders.
                    "rowBuffer": 0,
                    # How many blocks to keep in the store. Default is no limit, so every requested block is kept.
                    "maxBlocksInCache": 1,
                    "rowSelection": "multiple",
                },
            )
        ]
    )

    operators = {
        "greaterThanOrEqual": "ge",
        "lessThanOrEqual": "le",
        "lessThan": "lt",
        "greaterThan": "gt",
        "notEqual": "ne",
        "equals": "eq",
    }

    def filterDf(df, data, col):
        if data["filterType"] == "date":
            crit1 = data["dateFrom"]
            crit1 = pd.Series(crit1).astype(df[col].dtype)[0]
            if "dateTo" in data:
                crit2 = data["dateTo"]
                crit2 = pd.Series(crit2).astype(df[col].dtype)[0]
        else:
            crit1 = data["filter"]
            crit1 = pd.Series(crit1).astype(df[col].dtype)[0]
            if "filterTo" in data:
                crit2 = data["filterTo"]
                crit2 = pd.Series(crit2).astype(df[col].dtype)[0]
        if data["type"] == "contains":
            df = df.loc[df[col].str.contains(crit1)]
        elif data["type"] == "notContains":
            df = df.loc[~df[col].str.contains(crit1)]
        elif data["type"] == "startsWith":
            df = df.loc[df[col].str.startswith(crit1)]
        elif data["type"] == "notStartsWith":
            df = df.loc[~df[col].str.startswith(crit1)]
        elif data["type"] == "endsWith":
            df = df.loc[df[col].str.endswith(crit1)]
        elif data["type"] == "notEndsWith":
            df = df.loc[~df[col].str.endswith(crit1)]
        elif data["type"] == "inRange":
            if data["filterType"] == "date":
                df = df.loc[df[col].astype("datetime64[ns]").between_time(crit1, crit2)]
            else:
                df = df.loc[df[col].between(crit1, crit2)]
        elif data["type"] == "blank":
            df = df.loc[df[col].isnull()]
        elif data["type"] == "notBlank":
            df = df.loc[~df[col].isnull()]
        else:
            df = df.loc[getattr(df[col], operators[data["type"]])(crit1)]
        return df

    @app.callback(Output("grid", "getRowsResponse"), Input("grid", "getRowsRequest"))
    def infinite_scroll(request):
        dff = df.copy()

        if request:
            if request["filterModel"]:
                fils = request["filterModel"]
                for k in fils:
                    try:
                        if "operator" in fils[k]:
                            if fils[k]["operator"] == "AND":
                                dff = filterDf(dff, fils[k]["condition1"], k)
                                dff = filterDf(dff, fils[k]["condition2"], k)
                            else:
                                dff1 = filterDf(dff, fils[k]["condition1"], k)
                                dff2 = filterDf(dff, fils[k]["condition2"], k)
                                dff = pd.concat([dff1, dff2])
                        else:
                            dff = filterDf(dff, fils[k], k)
                    except:
                        pass
                dff = dff

            if request["sortModel"]:
                sorting = []
                asc = []
                for sort in request["sortModel"]:
                    sorting.append(sort["colId"])
                    if sort["sort"] == "asc":
                        asc.append(True)
                    else:
                        asc.append(False)
                dff = dff.sort_values(by=sorting, ascending=asc)

            lines = len(dff.index)
            if lines == 0:
                lines = 1

            partial = dff.iloc[request["startRow"] : request["endRow"]]
            return {"rowData": partial.to_dict("records"), "rowCount": lines}

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")
    grid.wait_for_cell_text(0, 0, "0")

    ## testing filtering with responses
    grid.set_filter(0, "1001")
    grid.wait_for_cell_text(0, 0, "1001")
    grid.set_filter(0, "")
    grid.wait_for_cell_text(0, 0, "0")

    ## testing sorting with responses
    grid.get_header_cell(0).click()
    grid.wait_for_cell_text(0, 0, "0")
    grid.get_header_cell(0).click()
    grid.wait_for_cell_text(0, 0, "9999")
    grid.get_header_cell(1).click()
    grid.wait_for_cell_text(1, 0, "3600")
    grid.get_header_cell(1).click()
    grid.wait_for_cell_text(0, 0, "7263")


def test_is002_infinite_scroll_styling(dash_duo):
    app = Dash(__name__)

    df = pd.read_csv(
        "https://raw.githubusercontent.com/plotly/datasets/master/ag-grid/olympic-winners.csv"
        # "data.csv"
    )

    crimson_palette = ['#fff5f0', '#fee0d2', '#fcbba1', '#fc9272', '#fb6a4a'][::-1]
    getRowStyle = {
        "styleConditions": [
            {
                "condition": "params.data['athlete'].includes('Michael')",
                "style": {
                    "backgroundColor": crimson_palette[0],
                    'color': 'whitesmoke'
                },
            },
            {
                "condition": "params.data['athlete'].includes('Ryan')",
                "style": {
                    "backgroundColor": crimson_palette[1],
                    # 'color': 'whitesmoke'
                },
            },
        ]
    }

    columnDefs = [
        # this row shows the row index, doesn't use any data from the row
        {
            "headerName": "ID",
            "maxWidth": 100,
            # it is important to have node.id here, so that when the id changes (which happens when the row is loaded)
            # then the cell is refreshed.
            "valueGetter": {"function": "params.node.id"},
        },
        {"field": "athlete", "minWidth": 150},
        {"field": "country", "minWidth": 150},
        {"field": "year"},
        {"field": "sport", "minWidth": 150},
        {"field": "total"},
    ]

    defaultColDef = {
        "flex": 1,
        "minWidth": 150,
        "sortable": False,
        "resizable": True,
    }

    app.layout = html.Div(
        [
            dag.AgGrid(
                id="infinite-row-no-sort",
                columnDefs=columnDefs,
                defaultColDef=defaultColDef,
                rowModelType="infinite",
                getRowStyle=getRowStyle,
            ),
            html.Button('scroll', id='scroll')
        ],
    )

    @app.callback(
        Output("infinite-row-no-sort", "getRowsResponse"),
        Input("infinite-row-no-sort", "getRowsRequest"),
    )
    def infinite_scroll(request):
        if request is None:
            return no_update
        partial = df.iloc[request["startRow"]: request["endRow"]]
        return {"rowData": partial.to_dict("records"), "rowCount": len(df.index)}

    @app.callback(
        Output("infinite-row-no-sort", "scrollTo"),
        Input('scroll', 'n_clicks'),
        prevent_initial_call=True
    )
    def scroll(n):
        return {'rowIndex': n*1000+n}

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "infinite-row-no-sort")
    grid.wait_for_cell_text(0, 0, "0")
    time.sleep(5)
    for x in range(8):
        dash_duo.find_element("#scroll").click()
        time.sleep(3)  # pausing to emulate separation because user inputs
    assert list(filter(lambda i: i.get("level") != "WARNING", dash_duo.get_logs())) == []


from dash import Dash, html, Input, Output, Patch, State
from dash_ag_grid import AgGrid
import plotly.express as px
import json
from dash.testing.wait import until
import pandas as pd
import pytest
import time

from . import utils


@pytest.fixture
def df():
    return pd.read_csv(
        "https://raw.githubusercontent.com/plotly/datasets/master/ag-grid/olympic-winners.csv"
    )


@pytest.fixture
def scroll_to_inputs():
    return [
        {"rowIndex": 100, "rowPosition": "bottom", "cell": (100, 0)},
        {"rowIndex": 100, "column": "bronze", "columnPosition": "end", "cell": (100, 8)},
        {"rowId": "Elizabeth Beisel12/8/2012", "column": "bronze", "columnPosition": "end", "rowPosition": "top", "cell": (200, 8)},
        {"rowIndex": 300, "column": "bronze", "columnPosition": "end", "cell": (300, 8)},
        {
            "rowIndex": 400,
            "rowId": "Ryan Bayley29/08/2004",
            "column": "athlete",
            "rowPosition": "bottom",
            "columnPosition": "start",
            "cell": (400, 0),
        },
        {
            "data": {
                "athlete": "Sabine Völker",
                "age": 28,
                "country": "Germany",
                "year": 2002,
                "date": "24/02/2002",
                "sport": "Speed Skating",
                "gold": 0,
                "silver": 2,
                "bronze": 1,
                "total": 3,
            },
            "column": "bronze",
            "columnPosition": "end",
            "cell": (100, 8),
        },
        {
            "rowIndex": 2000,
            "rowId": "Elizabeth Beisel12/8/2012",
            "data": {
                "athlete": "Elizabeth Beisel",
                "age": 19,
                "country": "United States",
                "year": 2012,
                "date": "12/8/2012",
                "sport": "Swimming",
                "gold": 0,
                "silver": 1,
                "bronze": 1,
                "total": 2,
            },
            "column": "age",
            "columnPosition": "start",
            "cell": (2000, 1),
        },
    ]


def test_st001_scroll_to(dash_duo, df, scroll_to_inputs):
    app = Dash()

    # basic columns definition with column defaults
    columnDefs = [{"field": c} for c in df.columns]

    app.layout = html.Div(
        [
            AgGrid(
                id="grid",
                columnDefs=columnDefs,
                rowData=df.to_dict("records"),
                defaultColDef={"resizable": True, "sortable": True, "filter": True},
                getRowId="params.data.athlete+params.data.date",
            ),
            html.Button(id="btn"),
            html.Button(id="btn_columnState"),
            html.Div(id="scrollTo-output"),
            html.Div(id="scrollTo-input"),
        ]
    )

    # Displays the prop scrollTo when it changes in the grid
    @app.callback(
        Output("scrollTo-output", "children"),
        Input("grid", "scrollTo"),
    )
    def display_scrollTo(scroll_to):
        if scroll_to:
            return json.dumps(scroll_to)

    @app.callback(
        Output('grid', 'columnState'),
        Input('btn_columnState', 'n_clicks'),
        State('grid', 'columnState'),
        prevent_initial_call=True
    )
    def reset_columnState(n, s):
        state = Patch()
        state[0]['width'] = s[0]['width'] - n+1
        return state

    # On click sets up a new value for scrollTo from the fixture scroll_to_inputs
    @app.callback(
        Output("grid", "scrollTo"),
        Input("btn", "n_clicks"),
        prevent_initial_call=True,
    )
    def update_scrollTo(n_clicks):
        return scroll_to_inputs[n_clicks - 1]

    dash_duo.driver.set_window_size(800, 600)  # Make window small enough to scroll things
    dash_duo.start_server(app)
    grid = utils.Grid(dash_duo, "grid")

    action = utils.ActionChains(dash_duo.driver)

    # Check that the grid has been loaded successfully
    until(lambda: "Michael Phelps" == grid.get_cell(0, 0).text, timeout=3)

    for i in range(len(scroll_to_inputs)):
        info = scroll_to_inputs[i]

        y, x = info["cell"]
        dash_duo.find_element("#btn").click()
        # removing as this no longer works due to resetting the value, [#313](https://github.com/plotly/dash-ag-grid/pull/313)
        # dash_duo.wait_for_text_to_equal("#scrollTo-output", json.dumps(info), timeout=5)
        until(lambda: grid.get_cell(y, x).is_displayed(), timeout=3)

        # row testing
        if "rowPosition" in info:
            if info["rowPosition"] == "bottom":
                assert not grid.cell_in_viewport(y + 1, x)
                assert grid.cell_in_viewport(y - 1, x)
            elif info["rowPosition"] == "middle":
                assert grid.cell_in_viewport(y + 1, x)
                assert grid.cell_in_viewport(y - 1, x)
            else:
                assert grid.cell_in_viewport(y + 1, x)
                assert not grid.cell_in_viewport(y - 1, x)
        elif "rowIndex" in info or "rowId" in info or "data" in info:
            assert grid.cell_in_viewport(y + 1, x)
            assert not grid.cell_in_viewport(y - 1, x)

        # column testing
        if "column" in info:
            if "columnPosition" in info:
                if info["columnPosition"] == "end":
                    if x + 1 < len(df.columns):
                        assert not grid.cell_in_viewport(y, x + 1)
                    if x - 1 >= 0:
                        assert grid.cell_in_viewport(y, x - 1)
                elif info["columnPosition"] == "middle":
                    if x + 1 < len(df.columns):
                        assert grid.cell_in_viewport(y, x + 1)
                    if x - 1 >= 0:
                        assert grid.cell_in_viewport(y, x - 1)
                else:
                    if x + 1 < len(df.columns):
                        assert grid.cell_in_viewport(y, x + 1)
                    if x - 1 >= 0:
                        assert not grid.cell_in_viewport(y, x - 1)
            else:
                if x + 1 < len(df.columns):
                    assert grid.cell_in_viewport(y, x + 1)
                if x - 1 >= 0:
                    assert not grid.cell_in_viewport(y, x - 1)

        # resets the grid
        dash_duo.driver.execute_script("""
            dash_ag_grid.getApi('grid').ensureIndexVisible(0);
            dash_ag_grid.getApi('grid').ensureColumnVisible('athlete');
         """)
        until(lambda: grid.get_cell(0, 0).is_displayed(), timeout=3)
        # make sure grid doesnt change upon double-click
        action.double_click(grid.get_cell(0, 0)).perform()
        time.sleep(1)
        until(lambda: grid.get_cell(0, 0).is_displayed(), timeout=3)
        # make sure scroll doesnt fire upon triggered reload by columnState
        dash_duo.find_element("#btn_columnState").click()
        time.sleep(1)
        until(lambda: grid.get_cell(0, 0).is_displayed(), timeout=3)



def test_st002_initial_scroll_to(dash_duo, df):
    app = Dash()

    # basic columns definition with column defaults
    columnDefs = [{"field": c} for c in df.columns]

    app.layout = html.Div(
        [
            AgGrid(
                id="grid",
                columnDefs=columnDefs,
                rowData=df.to_dict("records"),
                defaultColDef={"resizable": True, "sortable": True, "filter": True},
                scrollTo={
                    "rowIndex": 2000,
                    "rowPosition": "top",
                    "column": "bronze",
                    "columnPosition": "end",
                }
            ),
            html.Button(id="btn_columnState"),
        ]
    )

    @app.callback(
        Output('grid', 'columnState'),
        Input('btn_columnState', 'n_clicks'),
        State('grid', 'columnState'),
        prevent_initial_call=True
    )
    def reset_columnState(n, s):
        state = Patch()
        state[0]['width'] = s[0]['width'] - n + 1
        return state

    dash_duo.driver.set_window_size(800, 600)  # Make window small enough to scroll things
    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")
    until(lambda: "1" == grid.get_cell(2000, 9).text, timeout=3)

    y = 2000
    x = 8

    until(
        lambda: grid.get_cell(y, x).is_displayed(),
        timeout=3,
    )

    # row testing
    assert grid.cell_in_viewport(y + 1, x)
    assert not grid.cell_in_viewport(y - 1, x)

    # column testing
    assert not grid.cell_in_viewport(y, x + 1)
    assert grid.cell_in_viewport(y, x - 1)

    action = utils.ActionChains(dash_duo.driver)

    # resets the grid
    dash_duo.driver.execute_script("""
                dash_ag_grid.getApi('grid').ensureIndexVisible(0);
                dash_ag_grid.getApi('grid').ensureColumnVisible('athlete');
             """)
    until(lambda: grid.get_cell(0, 0).is_displayed(), timeout=3)
    # make sure grid doesnt change upon double-click
    action.double_click(grid.get_cell(0, 0)).perform()
    time.sleep(1)
    until(lambda: grid.get_cell(0, 0).is_displayed(), timeout=3)
    # make sure scroll doesnt fire upon triggered reload by columnState
    dash_duo.find_element("#btn_columnState").click()
    time.sleep(1)
    until(lambda: grid.get_cell(0, 0).is_displayed(), timeout=3)



# AUTO GENERATED FILE - DO NOT EDIT

import typing  # noqa: F401
from typing_extensions import TypedDict, NotRequired, Literal # noqa: F401
from dash.development.base_component import Component, _explicitize_args

ComponentSingleType = typing.Union[str, int, float, Component, None]
ComponentType = typing.Union[
    ComponentSingleType,
    typing.Sequence[ComponentSingleType],
]

NumberType = typing.Union[
    typing.SupportsFloat, typing.SupportsInt, typing.SupportsComplex
]


class AgGrid(Component):
    """An AgGrid component.
Dash interface to AG Grid, a powerful tabular data component.

Keyword arguments:

- id (string; optional):
    The ID used to identify this component in Dash callbacks.

- cellClicked (dict; optional):
    Cell is clicked.

    `cellClicked` is a dict with keys:

    - value (boolean | number | string | dict | list; optional):
        value of the clicked cell.

    - colId (boolean | number | string | dict | list; optional):
        column where the cell was clicked.

    - rowIndex (number; optional):
        rowIndex, typically a row number.

    - rowId (boolean | number | string | dict | list; optional):
        Row Id from the grid, this could be a number automatically, or
        set via getRowId.

    - timestamp (boolean | number | string | dict | list; optional):
        timestamp of last action.

- cellDoubleClicked (dict; optional):
    Cell is double clicked.

    `cellDoubleClicked` is a dict with keys:

    - value (boolean | number | string | dict | list; optional):
        value of the double-clicked cell.

    - colId (boolean | number | string | dict | list; optional):
        column where the cell was double-clicked.

    - rowIndex (number; optional):
        rowIndex, typically a row number.

    - rowId (boolean | number | string | dict | list; optional):
        Row Id from the grid, this could be a number automatically, or
        set via getRowId.

    - timestamp (boolean | number | string | dict | list; optional):
        timestamp of last action.

- cellRendererData (dict; optional):
    Special prop to allow feedback from cell renderer to the grid.

    `cellRendererData` is a dict with keys:

    - value (boolean | number | string | dict | list; optional):
        Value set from the function.

    - colId (string; optional):
        Column ID from where the event was fired.

    - rowIndex (number; optional):
        Row Index from the grid, this is associated with the row
        count.

    - rowId (boolean | number | string | dict | list; optional):
        Row Id from the grid, this could be a number automatically, or
        set via getRowId.

    - timestamp (boolean | number | string | dict | list; optional):
        Timestamp of when the event was fired.

- cellValueChanged (list of dicts; optional):
    Value has changed after editing.

    `cellValueChanged` is a list of dicts with keys:

    - rowIndex (number; optional):
        rowIndex, typically a row number.

    - rowId (boolean | number | string | dict | list; optional):
        Row Id from the grid, this could be a number automatically, or
        set via getRowId.

    - data (dict; optional):
        data, data object from the row.

    - oldValue (boolean | number | string | dict | list; optional):
        old value of the cell.

    - newValue (boolean | number | string | dict | list; optional):
        new value of the cell.

    - colId (boolean | number | string | dict | list; optional):
        column where the cell was changed.

    - timestamp (boolean | number | string | dict | list; optional):
        Timestamp of when the event was fired.

- className (string; default ''):
    The class for the ag-grid.  Can specify the ag-grid theme here.

- columnDefs (list of dicts; optional):
    Array of Column Definitions.

- columnSize (a value equal to: 'sizeToFit', 'autoSize', 'responsiveSizeToFit', null; optional):
    Size the columns autoSize changes the column sizes to fit the
    column's content, sizeToFit changes the column sizes to fit the
    width of the table responsiveSizeToFit changes the column sizes to
    fit the width of the table and also resizing upon grid or column
    changes and None bypasses the altering of the column widths.

- columnSizeOptions (dict; optional):
    Options to customize the columnSize operation. autoSize calls
    either autoSizeColumns or autoSizeAllColumns, see:
    https://www.ag-grid.com/react-data-grid/column-sizing/#autosize-column-api,
    and sizeToFit and responsiveSizeToFit call sizeColumnsToFit, see:
    https://www.ag-grid.com/react-data-grid/column-sizing/#size-columns-to-fit.

    `columnSizeOptions` is a dict with keys:

    - columnLimits (list of dicts; optional):
        for (responsive)sizeToFit: per-column minimum and maximum
        width, in pixels.

        `columnLimits` is a list of dicts with keys:

        - key (string; optional)

        - minWidth (number; optional)

        - maxWidth (number; optional)

    - defaultMinWidth (number; optional):
        for (responsive)sizeToFit: default minimum width, in pixels,
        if not overridden by columnLimits.

    - defaultMaxWidth (number; optional):
        for (responsive)sizeToFit: default maximum width, in pixels,
        if not overridden by columnLimits.

    - keys (list of strings; optional):
        for autoSize: list of column keys to autosize. If omitted, all
        columns will be autosized.

    - skipHeader (boolean; optional):
        for autoSize: If skipHeader=True, the header won't be included
        when calculating the column widths. default: False.

- columnState (list; optional):
    Current state of the columns.

- csvExportParams (dict; optional):
    Object with properties to pass to the exportDataAsCsv() method.

    `csvExportParams` is a dict with keys:

    - columnSeparator (string; optional):
        Delimiter to insert between cell values.

    - suppressQuotes (boolean; optional):
        Pass True to insert the value into the CSV file without
        escaping. In this case it is your responsibility to ensure
        that no cells contain the columnSeparator character.

    - prependContent (string; optional):
        Content to put at the top of the file export. A 2D array of
        CsvCell objects.

    - appendContent (string; optional):
        Content to put at the bottom of the file export.

    - allColumns (boolean; optional):
        If True, all columns will be exported in the order they appear
        in the columnDefs.

    - columnKeys (list of strings; optional):
        Provide a list (an array) of column keys or Column objects if
        you want to export specific columns.

    - fileName (string; optional):
        String to use as the file name.

    - onlySelected (boolean; optional):
        Export only selected rows.

    - onlySelectedAllPages (boolean; optional):
        Only export selected rows including other pages (only makes
        sense when using pagination).

    - skipColumnGroupHeaders (boolean; optional):
        Set to True to skip include header column groups.

    - skipColumnHeaders (boolean; optional):
        Set to True if you don't want to export column headers.

    - skipRowGroups (boolean; optional):
        Set to True to skip row group headers if grouping rows. Only
        relevant when grouping rows.

    - skipPinnedTop (boolean; optional):
        Set to True to suppress exporting rows pinned to the top of
        the grid.

    - skipPinnedBottom (boolean; optional):
        Set to True to suppress exporting rows pinned to the bottom of
        the grid.

- dangerously_allow_code (boolean; default False):
    Allow strings containing JavaScript code or HTML in certain props.
    If your app stores Dash layouts for later retrieval this is
    dangerous because it can lead to cross-site-scripting attacks.

- dashGridOptions (dict; optional):
    Other ag-grid options.

- dashRenderType (string; optional):
    dashRenderType to determine why grid is rendering.

- defaultColDef (dict; optional):
    A default column definition.

- deleteSelectedRows (boolean; optional):
    If True, the internal method deleteSelectedRows() will be called.

- deselectAll (boolean; default False):
    If True, the internal method deselectAll() will be called.

- detailCellRendererParams (dict; optional):
    Specifies the params to be used by the default detail Cell
    Renderer. See Detail Grids.

    `detailCellRendererParams` is a dict with keys:

    - detailGridOptions (boolean | number | string | dict | list; optional):
        Grid options for detail grid in master-detail view.

    - detailColName (string; optional):
        Column name where detail grid data is located in main dataset,
        for master-detail view.

    - suppressCallback (boolean; optional):
        Default: True. If True, suppresses the Dash callback in favor
        of using the data embedded in rowData at the given
        detailColName.

- enableEnterpriseModules (boolean; default False):
    If True, enable ag-grid Enterprise modules. Recommended to use
    with licenseKey.

- eventData (dict; optional):
    Object of Event Data from the grid, based upon when a user
    triggered an event with an event listener.

    `eventData` is a dict with keys:

    - data (boolean | number | string | dict | list; optional):
        Data that the developers passes back.

    - timestamp (boolean | number | string | dict | list; optional):
        Timestamp of when the event was fired.

- eventListeners (dict with strings as keys and values of type list; optional):
    Object of Eventlisteners to add upon grid ready. These listeners
    are only added upon grid ready. To add or remove an event listener
    after this point, please utilize the `getApi` or `getApiAsync`
    methods.

- exportDataAsCsv (boolean; default False):
    If True, the internal method exportDataAsCsv() will be called.

- filterModel (dict; optional):
    If filtering client-side rowModel, what the filter model is.
    Passing a model back to this prop will apply it to the grid.

- getDetailRequest (dict; optional):
    Request from Dash AgGrid when suppressCallback is disabled and a
    user opens a row with a detail grid.

    `getDetailRequest` is a dict with keys:

    - data (boolean | number | string | dict | list; optional):
        Details about the row that was opened.

    - requestTime (boolean | number | string | dict | list; optional):
        Datetime representing when the grid was requested.

- getDetailResponse (list of dicts; optional):
    RowData to populate the detail grid when callbacks are used to
    populate.

- getRowId (string; optional):
    This is required for change detection in rowData.

- getRowStyle (dict; optional):
    Object used to perform the row styling. See AG-Grid Row Style.

    `getRowStyle` is a dict with keys:

    - styleConditions (list of dicts; optional)

        `styleConditions` is a list of dicts with keys:

        - condition (string; required)

        - style (dict; required)

    - defaultStyle (dict; optional)

- getRowsRequest (dict; optional):
    Infinite Scroll, Datasource interface See
    https://www.ag-grid.com/react-grid/infinite-scrolling/#datasource-interface.

    `getRowsRequest` is a dict with keys:

    - startRow (number; optional):
        The first row index to get.

    - endRow (number; optional):
        The first row index to NOT get.

    - sortModel (list of dicts; optional):
        If sorting, what the sort model is.

    - filterModel (dict; optional):
        If filtering, what the filter model is.

    - context (boolean | number | string | dict | list; optional):
        The grid context object.

    - successCallback (optional):
        Callback to call when the request is successful.

    - failCallback (optional):
        Callback to call when the request fails.

- getRowsResponse (dict; optional):
    Serverside model data response object. See
    https://www.ag-grid.com/react-grid/server-side-model-datasource/.

    `getRowsResponse` is a dict with keys:

    - rowData (list of dicts; optional):
        Data retreived from the server.

    - rowCount (number; optional):
        Current row count, if known.

    - storeInfo (boolean | number | string | dict | list; optional):
        Any extra info for the grid to associate with this load.

- licenseKey (string; optional):
    License key for ag-grid enterprise. If using Enterprise modules,
    enableEnterpriseModules must also be True.

- masterDetail (boolean; optional):
    Used to enable Master Detail. See Enabling Master Detail. Default
    Value: False.

- paginationGoTo (a value equal to: 'first', 'last', 'next', 'previous', null | number; optional):
    If in pagination mode, this will navigate to: ['next', 'previous',
    'last', 'first', number]
    https://www.ag-grid.com/react-data-grid/grid-api/#reference-pagination.

- paginationInfo (dict; optional):
    If in pagination mode, this will be populated with info from the
    pagination API:
    https://www.ag-grid.com/react-data-grid/grid-api/#reference-pagination.

    `paginationInfo` is a dict with keys:

    - isLastPageFound (boolean; optional)

    - pageSize (number; optional)

    - currentPage (number; optional)

    - totalPages (number; optional)

    - rowCount (number; optional)

- persisted_props (list of strings; default ['selectedRows']):
    Properties whose user interactions will persist after refreshing
    the component or the page.

- persistence (boolean | string | number; optional):
    Used to allow user interactions in this component to be persisted
    when the component - or the page - is refreshed. If `persisted` is
    truthy and hasn't changed from its previous value, a `value` that
    the user has changed while using the app will keep that change, as
    long as the new `value` also matches what was given originally.
    Used in conjunction with `persistence_type`.

- persistence_type (a value equal to: 'local', 'session', 'memory'; default 'local'):
    Where persisted user changes will be stored: memory: only kept in
    memory, reset on page refresh. local: window.localStorage, data is
    kept after the browser quit. session: window.sessionStorage, data
    is cleared once the browser quit.

- resetColumnState (boolean; default False):
    If True, the internal method resetColumnState() will be called.

- rowClass (string; optional):
    The class to give a particular row. See Row Class.

- rowClassRules (dict; optional):
    Rules which can be applied to include certain CSS classes. See Row
    Class Rules.

- rowData (list of dicts; optional):
    (Client-Side Row Model only) Set the data to be displayed as rows
    in the grid.

- rowModelType (a value equal to: 'clientSide', 'infinite', 'viewport', 'serverSide'; default 'clientSide'):
    Sets the Row Model type. Default Value: 'clientSide'.

- rowStyle (dict; optional):
    The style to give a particular row. See Row Style.

- rowTransaction (dict; optional):
    If True, the internal method rowTransaction() will be used, if
    async provided as False, applyTransaction() will be called, else
    applyTransactionAsync().

    `rowTransaction` is a dict with keys:

    - async (boolean; optional)

    - add (list; optional)

    - update (list; optional)

    - remove (list; optional)

    - addIndex (number; optional)

- scrollTo (dict; optional):
    Scrolls to a specific position.

    `scrollTo` is a dict with keys:

    - rowIndex (number; optional):
        rowIndex, typically a row number.

    - rowId (string; optional):
        Id of the row to scroll to.

    - data (dict; optional):
        Data of the row to scroll to.

    - rowPosition (a value equal to: 'top', 'bottom', 'middle'; optional):
        Position of the row in the grid after scrolling. Default
        `top`.

    - column (string; optional):
        Column to scroll to, must be equal to one `field` in
        `columnDefs`.

    - columnPosition (a value equal to: 'auto', 'start', 'middle', 'end'; optional):
        Position of the column in the grid after scrolling. Default
        `auto`.

- selectAll (dict; default False):
    Set to True to cause all rows to be selected, Or pass an object of
    options for which rows to select. Currently supports `filtered`,
    set to True to only select filtered rows.

    `selectAll` is a boolean | dict with keys:

    - filtered (boolean; optional)

- selectedRows (list of dicts; optional):
    The actively selected rows from the grid (may include filtered
    rows) Can take one of three forms: (1) an array of row objects -
    if you have defined `getRowId`, you only need the fields it uses.
    (2) an object containing `function` with a function string - see:
    https://www.ag-grid.com/react-data-grid/row-selection/#example-using-foreachnode
    (selectAllAmerican function) (3) an object containing `ids` with a
    list of row IDs.

    `selectedRows` is a list of dicts | dict with keys:

    - function (string; required)

      Or dict with keys:

    - ids (list of strings; required)

- suppressDragLeaveHidesColumns (boolean; default True):
    If True, when you drag a column out of the grid (e.g. to the group
    zone) the column is not hidden.

- updateColumnState (boolean; default False):
    If True, the internal method updateColumnState() will be called.

- virtualRowData (list of dicts; optional):
    The rowData in the grid after inline filters are applied."""
    _children_props: typing.List[str] = []
    _base_nodes = ['children']
    _namespace = 'dash_ag_grid'
    _type = 'AgGrid'
    SelectAll = TypedDict(
        "SelectAll",
            {
            "filtered": NotRequired[bool]
        }
    )

    RowTransaction = TypedDict(
        "RowTransaction",
            {
            "async": NotRequired[bool],
            "add": NotRequired[typing.Sequence],
            "update": NotRequired[typing.Sequence],
            "remove": NotRequired[typing.Sequence],
            "addIndex": NotRequired[NumberType]
        }
    )

    CsvExportParams = TypedDict(
        "CsvExportParams",
            {
            "columnSeparator": NotRequired[str],
            "suppressQuotes": NotRequired[bool],
            "prependContent": NotRequired[str],
            "appendContent": NotRequired[str],
            "allColumns": NotRequired[bool],
            "columnKeys": NotRequired[typing.Sequence[str]],
            "fileName": NotRequired[str],
            "onlySelected": NotRequired[bool],
            "onlySelectedAllPages": NotRequired[bool],
            "skipColumnGroupHeaders": NotRequired[bool],
            "skipColumnHeaders": NotRequired[bool],
            "skipRowGroups": NotRequired[bool],
            "skipPinnedTop": NotRequired[bool],
            "skipPinnedBottom": NotRequired[bool]
        }
    )

    ColumnSizeOptionsColumnLimits = TypedDict(
        "ColumnSizeOptionsColumnLimits",
            {
            "key": NotRequired[str],
            "minWidth": NotRequired[NumberType],
            "maxWidth": NotRequired[NumberType]
        }
    )

    ColumnSizeOptions = TypedDict(
        "ColumnSizeOptions",
            {
            "columnLimits": NotRequired[typing.Sequence["ColumnSizeOptionsColumnLimits"]],
            "defaultMinWidth": NotRequired[NumberType],
            "defaultMaxWidth": NotRequired[NumberType],
            "keys": NotRequired[typing.Sequence[str]],
            "skipHeader": NotRequired[bool]
        }
    )

    GetRowStyleStyleConditions = TypedDict(
        "GetRowStyleStyleConditions",
            {
            "condition": str,
            "style": dict
        }
    )

    GetRowStyle = TypedDict(
        "GetRowStyle",
            {
            "styleConditions": NotRequired[typing.Sequence["GetRowStyleStyleConditions"]],
            "defaultStyle": NotRequired[dict]
        }
    )

    GetRowsRequest = TypedDict(
        "GetRowsRequest",
            {
            "startRow": NotRequired[NumberType],
            "endRow": NotRequired[NumberType],
            "sortModel": NotRequired[typing.Sequence[dict]],
            "filterModel": NotRequired[dict],
            "context": NotRequired[typing.Any],
            "successCallback": NotRequired[typing.Any],
            "failCallback": NotRequired[typing.Any]
        }
    )

    PaginationInfo = TypedDict(
        "PaginationInfo",
            {
            "isLastPageFound": NotRequired[bool],
            "pageSize": NotRequired[NumberType],
            "currentPage": NotRequired[NumberType],
            "totalPages": NotRequired[NumberType],
            "rowCount": NotRequired[NumberType]
        }
    )

    GetDetailRequest = TypedDict(
        "GetDetailRequest",
            {
            "data": NotRequired[typing.Any],
            "requestTime": NotRequired[typing.Any]
        }
    )

    CellRendererData = TypedDict(
        "CellRendererData",
            {
            "value": NotRequired[typing.Any],
            "colId": NotRequired[str],
            "rowIndex": NotRequired[NumberType],
            "rowId": NotRequired[typing.Any],
            "timestamp": NotRequired[typing.Any]
        }
    )

    GetRowsResponse = TypedDict(
        "GetRowsResponse",
            {
            "rowData": NotRequired[typing.Sequence[dict]],
            "rowCount": NotRequired[NumberType],
            "storeInfo": NotRequired[typing.Any]
        }
    )

    ScrollTo = TypedDict(
        "ScrollTo",
            {
            "rowIndex": NotRequired[NumberType],
            "rowId": NotRequired[str],
            "data": NotRequired[dict],
            "rowPosition": NotRequired[Literal["top", "bottom", "middle"]],
            "column": NotRequired[str],
            "columnPosition": NotRequired[Literal["auto", "start", "middle", "end"]]
        }
    )

    EventData = TypedDict(
        "EventData",
            {
            "data": NotRequired[typing.Any],
            "timestamp": NotRequired[typing.Any]
        }
    )

    DetailCellRendererParams = TypedDict(
        "DetailCellRendererParams",
            {
            "detailGridOptions": NotRequired[typing.Any],
            "detailColName": NotRequired[str],
            "suppressCallback": NotRequired[bool]
        }
    )

    CellClicked = TypedDict(
        "CellClicked",
            {
            "value": NotRequired[typing.Any],
            "colId": NotRequired[typing.Any],
            "rowIndex": NotRequired[NumberType],
            "rowId": NotRequired[typing.Any],
            "timestamp": NotRequired[typing.Any]
        }
    )

    CellDoubleClicked = TypedDict(
        "CellDoubleClicked",
            {
            "value": NotRequired[typing.Any],
            "colId": NotRequired[typing.Any],
            "rowIndex": NotRequired[NumberType],
            "rowId": NotRequired[typing.Any],
            "timestamp": NotRequired[typing.Any]
        }
    )

    SelectedRows = TypedDict(
        "SelectedRows",
            {
            "ids": typing.Sequence[str]
        }
    )

    CellValueChanged = TypedDict(
        "CellValueChanged",
            {
            "rowIndex": NotRequired[NumberType],
            "rowId": NotRequired[typing.Any],
            "data": NotRequired[dict],
            "oldValue": NotRequired[typing.Any],
            "newValue": NotRequired[typing.Any],
            "colId": NotRequired[typing.Any],
            "timestamp": NotRequired[typing.Any]
        }
    )


    def __init__(
        self,
        id: typing.Optional[typing.Union[str, dict]] = None,
        style: typing.Optional[typing.Any] = None,
        className: typing.Optional[str] = None,
        persistence: typing.Optional[typing.Union[bool, str, NumberType]] = None,
        persisted_props: typing.Optional[typing.Sequence[str]] = None,
        persistence_type: typing.Optional[Literal["local", "session", "memory"]] = None,
        dangerously_allow_code: typing.Optional[bool] = None,
        resetColumnState: typing.Optional[bool] = None,
        exportDataAsCsv: typing.Optional[bool] = None,
        selectAll: typing.Optional[typing.Union[bool, "SelectAll"]] = None,
        deselectAll: typing.Optional[bool] = None,
        updateColumnState: typing.Optional[bool] = None,
        deleteSelectedRows: typing.Optional[bool] = None,
        rowTransaction: typing.Optional["RowTransaction"] = None,
        getRowId: typing.Optional[str] = None,
        columnState: typing.Optional[typing.Sequence] = None,
        csvExportParams: typing.Optional["CsvExportParams"] = None,
        columnSize: typing.Optional[Literal["sizeToFit", "autoSize", "responsiveSizeToFit", None]] = None,
        columnSizeOptions: typing.Optional["ColumnSizeOptions"] = None,
        getRowStyle: typing.Optional["GetRowStyle"] = None,
        getRowsRequest: typing.Optional["GetRowsRequest"] = None,
        paginationInfo: typing.Optional["PaginationInfo"] = None,
        paginationGoTo: typing.Optional[typing.Union[Literal["first", "last", "next", "previous", None], NumberType]] = None,
        filterModel: typing.Optional[dict] = None,
        getDetailRequest: typing.Optional["GetDetailRequest"] = None,
        getDetailResponse: typing.Optional[typing.Sequence[dict]] = None,
        cellRendererData: typing.Optional["CellRendererData"] = None,
        getRowsResponse: typing.Optional["GetRowsResponse"] = None,
        licenseKey: typing.Optional[str] = None,
        enableEnterpriseModules: typing.Optional[bool] = None,
        virtualRowData: typing.Optional[typing.Sequence[dict]] = None,
        scrollTo: typing.Optional["ScrollTo"] = None,
        eventListeners: typing.Optional[typing.Dict[typing.Union[str, float, int], typing.Sequence]] = None,
        eventData: typing.Optional["EventData"] = None,
        columnDefs: typing.Optional[typing.Sequence[dict]] = None,
        defaultColDef: typing.Optional[dict] = None,
        rowModelType: typing.Optional[Literal["clientSide", "infinite", "viewport", "serverSide"]] = None,
        rowData: typing.Optional[typing.Sequence[dict]] = None,
        masterDetail: typing.Optional[bool] = None,
        detailCellRendererParams: typing.Optional["DetailCellRendererParams"] = None,
        rowStyle: typing.Optional[dict] = None,
        rowClass: typing.Optional[str] = None,
        rowClassRules: typing.Optional[dict] = None,
        suppressDragLeaveHidesColumns: typing.Optional[bool] = None,
        cellClicked: typing.Optional["CellClicked"] = None,
        cellDoubleClicked: typing.Optional["CellDoubleClicked"] = None,
        selectedRows: typing.Optional[typing.Union[typing.Sequence[dict], "SelectedRows"]] = None,
        cellValueChanged: typing.Optional[typing.Sequence["CellValueChanged"]] = None,
        dashGridOptions: typing.Optional[dict] = None,
        dashRenderType: typing.Optional[str] = None,
        **kwargs
    ):
        self._prop_names = ['id', 'cellClicked', 'cellDoubleClicked', 'cellRendererData', 'cellValueChanged', 'className', 'columnDefs', 'columnSize', 'columnSizeOptions', 'columnState', 'csvExportParams', 'dangerously_allow_code', 'dashGridOptions', 'dashRenderType', 'defaultColDef', 'deleteSelectedRows', 'deselectAll', 'detailCellRendererParams', 'enableEnterpriseModules', 'eventData', 'eventListeners', 'exportDataAsCsv', 'filterModel', 'getDetailRequest', 'getDetailResponse', 'getRowId', 'getRowStyle', 'getRowsRequest', 'getRowsResponse', 'licenseKey', 'masterDetail', 'paginationGoTo', 'paginationInfo', 'persisted_props', 'persistence', 'persistence_type', 'resetColumnState', 'rowClass', 'rowClassRules', 'rowData', 'rowModelType', 'rowStyle', 'rowTransaction', 'scrollTo', 'selectAll', 'selectedRows', 'style', 'suppressDragLeaveHidesColumns', 'updateColumnState', 'virtualRowData']
        self._valid_wildcard_attributes =            []
        self.available_properties = ['id', 'cellClicked', 'cellDoubleClicked', 'cellRendererData', 'cellValueChanged', 'className', 'columnDefs', 'columnSize', 'columnSizeOptions', 'columnState', 'csvExportParams', 'dangerously_allow_code', 'dashGridOptions', 'dashRenderType', 'defaultColDef', 'deleteSelectedRows', 'deselectAll', 'detailCellRendererParams', 'enableEnterpriseModules', 'eventData', 'eventListeners', 'exportDataAsCsv', 'filterModel', 'getDetailRequest', 'getDetailResponse', 'getRowId', 'getRowStyle', 'getRowsRequest', 'getRowsResponse', 'licenseKey', 'masterDetail', 'paginationGoTo', 'paginationInfo', 'persisted_props', 'persistence', 'persistence_type', 'resetColumnState', 'rowClass', 'rowClassRules', 'rowData', 'rowModelType', 'rowStyle', 'rowTransaction', 'scrollTo', 'selectAll', 'selectedRows', 'style', 'suppressDragLeaveHidesColumns', 'updateColumnState', 'virtualRowData']
        self.available_wildcard_properties =            []
        _explicit_args = kwargs.pop('_explicit_args')
        _locals = locals()
        _locals.update(kwargs)  # For wildcard attrs and excess named props
        args = {k: _locals[k] for k in _explicit_args}

        super(AgGrid, self).__init__(**args)

setattr(AgGrid, "__init__", _explicitize_args(AgGrid.__init__))



from dash import Dash, html, Output, Input, no_update, ALL, ctx, Patch
import dash_ag_grid as dag
import plotly.express as px
import pandas as pd
import json
from dash.testing.wait import until
import time

from . import utils


def test_pa001_pagination(dash_duo):
    app = Dash(__name__)

    df = pd.read_csv(
        "https://raw.githubusercontent.com/plotly/datasets/master/ag-grid/olympic-winners.csv"
    )

    opts = ["first", "next", "previous", "last", 0, 40]

    # basic columns definition with column defaults
    columnDefs = [
        {"field": "country"},
        {"field": "year"},
        {"field": "athlete"},
        {"field": "age"},
        {"field": "date"},
        {"field": "sport"},
        {"field": "total"},
    ]

    app.layout = html.Div(
        [
            dag.AgGrid(
                id="grid",
                columnDefs=columnDefs,
                rowData=df.to_dict("records"),
                columnSize="sizeToFit",
                defaultColDef={"resizable": True, "sortable": True, "filter": True},
                dashGridOptions={"pagination": True},
                paginationGoTo=5,
            ),
            html.Div(id="grid-info"),

        ]
        + [html.Button(id={"type": "nav", "index": x}, children=x) for x in opts]
        + [html.Button(id="changeSize", children="changeSize")],
        style={"margin": 20},
    )

    @app.callback(Output("grid", "dashGridOptions"), Input("changeSize", "n_clicks"))
    def updatePageSize(n):
        if n:
            opts = Patch()
            opts["paginationPageSize"] = 10 * n
            return opts
        return no_update

    @app.callback(Output("grid-info", "children"), Input("grid", "paginationInfo"))
    def update_height(h):
        if h:
            return json.dumps(h)
        return no_update

    @app.callback(
        Output("grid", "paginationGoTo"),
        Input({"type": "nav", "index": ALL}, "n_clicks"),
    )
    def updatePage(_):
        if ctx.triggered:
            return ctx.triggered_id["index"]
        return no_update

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    time.sleep(1)  # wait for the grid to load
    until(lambda: "Australia" == grid.get_cell(500, 0).text, timeout=3)
    oldValue = '{"isLastPageFound": true, "pageSize": 100, "currentPage": 5, "totalPages": 87, "rowCount": 8618}'
    until(lambda: oldValue == dash_duo.find_element("#grid-info").text, timeout=3)
    oldValue = dash_duo.find_element("#grid-info").text
    dash_duo.find_element('.ag-paging-button[aria-label="Last Page"]').click()
    until(lambda: oldValue != dash_duo.find_element("#grid-info").text, timeout=3)
    oldValue = dash_duo.find_element("#grid-info").text

    btns = dash_duo.find_elements("button")
    for x in range(len(btns)):
        btns[x].click()
        until(lambda: oldValue != dash_duo.find_element("#grid-info").text, timeout=3)
        oldValue = dash_duo.find_element("#grid-info").text

    assert (
        oldValue == '{"isLastPageFound": true, "pageSize": 10, '
        '"currentPage": 0, "totalPages": 862, '
        '"rowCount": 8618}'
    )



import dash_ag_grid as dag
from dash import Dash, html, dcc
from . import utils
import time
import dash_mantine_components


def test_pp001_popupParent(dash_duo):
    app = Dash(__name__)

    app.layout = html.Div(
        [
            dag.AgGrid(
                rowData=[{'cities': ''}, {'cities': 'dallas'}, {'cities': 'houston'}],
                columnDefs=[{'field': 'cities', 'cellEditor': {'function': 'DMC_Select'},
                             'cellEditorParams': {'options': ['dallas', 'houston']},
                             'cellEditorPopup': True, 'editable': True}],
                style={'resize': 'both', 'overflow': 'auto'},
                dashGridOptions={'popupParent': {'function': 'setBody()'}},
                id='grid'
            )
        ],

    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    grid.wait_for_cell_text(0, 0, "")

    ### testing animations
    action = utils.ActionChains(dash_duo.driver)
    action.double_click(grid.get_cell(0, 0)).perform()

    dash_duo.find_element('body > .ag-popup .mantine-Select-input')



import dash_ag_grid as dag
from dash import Dash, html, dcc, Input, Output, State
from . import utils
from dash.testing.wait import until
import pandas as pd
import json


def test_rg001_row_groupings(dash_duo):
    app = Dash(__name__)

    df = pd.read_csv(
        "https://raw.githubusercontent.com/plotly/datasets/master/ag-grid/olympic-winners.csv"
    )

    columnDefs = [
        {"field": "country", "enableRowGroup": True, "rowGroup": True, "hide": True},
        {"field": "year", "enableRowGroup": True, "rowGroup": True, "hide": True},
        {"field": "sport", "enableRowGroup": True},
        {"field": "gold"},
        {"field": "silver"},
        {"field": "bronze"},
    ]

    app.layout = html.Div(
        [
            dag.AgGrid(
                columnDefs=columnDefs,
                rowData=df.to_dict("records")[:20],
                columnSize="sizeToFit",
                defaultColDef={"resizable": True, "sortable": True, "filter": True},
                enableEnterpriseModules=True,
                dashGridOptions={"rowGroupPanelShow": "always"},
                id="grid",
            ),
            html.Div(id="columnState"),
        ],
        style={"margin": 20},
    )

    @app.callback(Output("columnState", "children"), Input("grid", "columnState"))
    def columnState(s):
        return json.dumps(s)

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "grid")

    until(lambda: "United States" in grid.get_cell(0, 0).text, timeout=3)

    ## grouped columns are hidden
    grid.add_column_drop(3)

    until(lambda: "Country" in dash_duo.find_element(".ag-column-drop").text, timeout=3)
    until(lambda: "Year" in dash_duo.find_element(".ag-column-drop").text, timeout=3)
    until(lambda: "Sport" in dash_duo.find_element(".ag-column-drop").text, timeout=3)
    assert ["Country", "Year", "Sport"] == dash_duo.find_element(
        ".ag-column-drop"
    ).text.split("\n")

    grid.drag_column_list(2, 0)
    assert ["Sport", "Country", "Year"] == dash_duo.find_element(
        ".ag-column-drop"
    ).text.split("\n")
    until(lambda: "Swimming" in grid.get_cell(0, 0).text, timeout=3)
    grid.get_cell_expandable(0, 0).click()
    until(lambda: "United States" in grid.get_cell(1, 0).text, timeout=3)

def test_rg002_row_groupings(dash_duo):
    app = Dash(__name__)

    data = {
        "Ireland": ["Dublin", "Galway", "Cork"],
        "UK": ["London", "Bristol", "Manchester", "Liverpool"],
        "USA": ["New York", "Boston", "L.A.", "San Fransisco", "Detroit"],
        "MiddleEarth": ["The Shire", "Rohan", "Rivendell", "Mordor"],
        "Midkemia": ["Darkmoor", "Crydee", "Elvandar", "LaMut", "Ylith"],
    }

    rowData = []
    for country, cities in data.items():
        for city in cities:
            rowData.append(
                {
                    "country": country,
                    "type": "Non Fiction"
                    if country in ["Ireland", "UK", "USA"]
                    else "Fiction",
                    "city": city,
                }
            )

    columnDefs = [
        # this column shows just the country group values, but has not group renderer, so there is no expand / collapse functionality
        {
            "headerName": "Country Group - No Renderer",
            "showRowGroup": "country",
        },
        # same as before, but we show all group values, again with no cell renderer
        {
            "headerName": "All Groups - No Renderer",
            "showRowGroup": True,
        },
        # add in a cell renderer
        {
            "headerName": "Group Renderer A",
            "showRowGroup": True,
            "cellRenderer": "agGroupCellRenderer",
        },
        # add in a field
        {
            "headerName": "Group Renderer B",
            "field": "city",
            "showRowGroup": True,
            "cellRenderer": "agGroupCellRenderer",
        },
        # add in a cell renderer params
        {
            "headerName": "Group Renderer C",
            "field": "city",
            "showRowGroup": True,
            "cellRenderer": "agGroupCellRenderer",
            "cellRendererParams": {
                "suppressCount": True,
                "checkbox": True,
                'innerRenderer': "SimpleCellRenderer",
                "suppressDoubleClickExpand": True,
                "suppressEnterExpand": True,
            },
        },
        {"headerName": "Type", "field": "type", "rowGroup": True},
        {"headerName": "Country", "field": "country", "rowGroup": True},
        {"headerName": "City", "field": "city", "editable": True},
    ]

    @app.callback(
        Output("grid-virtualRowData-info", "children"),
        Input("group-cell-renderer-example", "cellValueChanged"),
        State("group-cell-renderer-example", "rowData"),
        State("group-cell-renderer-example", "virtualRowData"),
        prevent_initial_call=True
    )
    def update(cell_changed, row_data, virtual_row_data):
        return f"{virtual_row_data[0]}"

    app.layout = html.Div(
        [
            dag.AgGrid(
                id="group-cell-renderer-example",
                columnDefs=columnDefs,
                rowData=rowData,
                columnSize="sizeToFit",
                defaultColDef={"resizable": True},
                enableEnterpriseModules=True,
                dangerously_allow_code=True,
                dashGridOptions={
                    "groupSelectsChildren": True,
                    "groupDisplayType": "custom",
                    "groupDefaultExpanded": 1,
                    "rowSelection": "multiple",
                }
            ),
            html.Div(id="grid-virtualRowData-info"),
        ],
        style={"margin": 20},
    )

    dash_duo.start_server(app)

    grid = utils.Grid(dash_duo, "group-cell-renderer-example")

    grid.wait_for_cell_text(1,1, 'Ireland')

    grid.get_cell_expandable(1,4).click()

    grid.wait_for_cell_text(2, 7, 'Dublin')

    grid.get_cell(2, 7).send_keys("texas")
    grid.get_cell(0, 0).click()
    grid.wait_for_cell_text(2, 7, 'texas')

    until(lambda: dash_duo.find_element('#grid-virtualRowData-info').text
                  == "{'country': 'Ireland', 'type': 'Non Fiction', 'city': 'texas'}",
          timeout=3)